unit PlotterUnit;

interface

uses
  Windows, SysUtils, Forms, Graphics, Controls, Classes, Math, IniFiles, JPEG;

const
  FigureConstraint = 120;
  AxesFormat = '%0.4g';
  GridCellSize = 45;
  InnerMargin = 1;
  AxesTextMargin = 2;
  TitleTextMargin = 2;
  DeltaFont = 1;
  EPS = 1e-12;
  SpaceChars = ' '#9#10#13',;';

type
  TArrayOfInteger = array of integer;
  TArrayOfDouble = array of double;

  TFigure = class
  private
    FCanvas: TCanvas;
    FRect: TRect;
    Id, N: integer;
    X, Y: array of TArrayOfDouble;

    clLine, clFill, clPoint: TArrayOfInteger;
    clAxis, clBack, clUnder, clGrid: TColor;
    LineWidth, PointSize: TArrayOfInteger;
    Margin, LeftMargin, BottomMargin: integer;
    ShowTitle, ShowXnums, ShowYnums, ShowXgrid, ShowYgrid, LogX, LogY: boolean;
    MinX, MinY, MaxX, MaxY, SMinX, SMinY, SMaxX, SMaxY: double;
    FontName, Title: string;    
    TitleVisible: boolean;
    TitleHeight: integer;
    FontSize: integer;

    procedure ClipCanvas(const Rect: TRect);
    procedure UnClipCanvas;
    function OrderWidth(const Order: integer): integer;
    function OrderHeight(const Order: integer): integer;
    procedure DrawOrder(const Order: integer; const P: TPoint);

    procedure SaveToBMP(const FileName: string; const Width, Height: integer);
    procedure SaveToJPG(const FileName: string; const Width, Height: integer);
    procedure SaveToEMF(const FileName: string; const Width, Height: integer);
  public
    property Canvas: TCanvas read FCanvas write FCanvas;

    constructor Create(const FigureNumber: integer);
    procedure Resize(const NewRect: TRect);
    procedure Draw;
    procedure SaveToFile(const FileName: string; const Width, Height: integer);
  end;

  TFigureWindow = class
  private
    FCanvas: TCanvas;
    FWidth, FHeight: integer;

    Ini: TMemIniFile;

    Cap: string;
    Figure: array of TFigure;

    function GetFigure(FigureNumber: integer): TFigure;
    procedure SetCanvas(const NewCanvas: TCanvas);
  public
    StartFig, EndFig, BegFig, WFig, HFig: integer;

    property Caption: string read Cap;
    property Figures[FigureNumber: integer]: TFigure read GetFigure;
    property Canvas: TCanvas read FCanvas write SetCanvas;
    property Width: integer read FWidth;
    property Height: integer read FHeight;
    procedure Resize(const NewWidth, NewHeight: integer);

    procedure HomeFigure;
    procedure EndFigure;
    function NextFigure(const One: boolean = true): boolean;
    function PrevFigure(const One: boolean = true): boolean;
    procedure MoreFigure(const Xs: boolean = true);
    function LessFigure(const Xs: boolean = true): boolean;

    constructor Create(const FileName: string);
    destructor Destroy; override;

    procedure Load(const ReadSizes: boolean = true);
    procedure Draw;
    function Hint(const X, Y: integer; const Precision: byte;
      var HintType: integer): string;
    procedure SaveToFile(const FileName: string);
    function WhatFigure(const X, Y: integer): integer;
    function GetRect(const FigureNumber: integer): TRect;
    function GetTitleRect(const FigureNumber: integer): TRect;
    function GetGraphRect(const FigureNumber: integer): TRect;
  end;

var
  OptionsFile, PlottoFile: string;

implementation

procedure Swap(var A, B: double);
var
  C: double;
begin
  C := A; A := B; B := C;
end;

procedure Ignore(const Str: string; var Idx: integer; const Sym: string);
begin
  while (Idx <= length(Str)) and (Pos(Str[Idx], Sym) <> 0) do Idx := Idx + 1;
end;

function StrTok(const Str: string; var Idx: integer;
  const Delim: string): string;
var
  i: integer;
begin
  Ignore(Str, Idx, Delim);
  i := Idx;
  repeat i := i + 1;
  until (i > length(Str)) or (pos(Str[i], Delim) <> 0);
  result := copy(Str, Idx, i-Idx);
  Idx := i;
end;

function InRect(const Rect: TRect; const X, Y: integer): boolean;
begin
  result := (X >= Rect.Left) and (X <= Rect.Right) and
    (Y >= Rect.Top) and (Y <= Rect.Bottom);
end;

// TFigure /////////////////////////////////////////////////////////////////////

constructor TFigure.Create(const FigureNumber: integer);
begin
  N := 0;
  Id := FigureNumber;
  setlength(X, 0, 0);
  setlength(Y, 0, 0);
end;

procedure TFigure.Resize(const NewRect: TRect);
var
  i, W, H, Order, grid_x, grid_y: integer;
  kX, kY, DivNum, gx, gy, q, q0: double;
  Str: string;
begin
  FRect := NewRect;

  W := FRect.right - FRect.left + 1 - 3*Margin - LeftMargin;
  H := FRect.bottom - FRect.top + 1 - 3*Margin - BottomMargin;
  if ShowTitle then
    Dec(H, TitleHeight + Margin div 2);

  if (SMaxX > NegInfinity) and (SMinX < Infinity) and
     (SMaxY > NegInfinity) and (SMinY < Infinity) then begin
    kX := abs(SMaxX - SMinX)/W;
    kY := abs(SMaxY - SMinY)/H;

    MinX := SMinX - kX*InnerMargin;
    MaxX := SMaxX + kX*InnerMargin;
    MinY := SMinY - kY*InnerMargin;
    MaxY := SMaxY + kY*InnerMargin;
  end;

  LeftMargin := Margin;
  BottomMargin := Margin;
  if (FCanvas <> nil) and (N > 0) then begin
    if ShowXnums then begin
      BottomMargin := -1;
      if LogX then begin
        Order := round(MaxX - MinX);
        DivNum := IntPower(10, Order);
        for i := round(MinX) to round(MaxX) do begin
          Str := Format(AxesFormat, [IntPower(10, i)/DivNum]);
          BottomMargin := max(BottomMargin, FCanvas.TextHeight(Str));
        end;
      end else begin
        grid_x := max(1, W div GridCellSize);
        gx := (MaxX - MinX)/grid_x;
        Order := round(log10(gx));
        gx := round(gx*IntPower(10, -Order+1))*IntPower(10, Order-1);
        DivNum := IntPower(10, Order);
        if gx > 0 then begin
          q0 := gx*int(MinX/gx);
          i := 0; q := q0;
          while q <= MaxX do begin
            if abs(q) < EPS then q := 0;
            Str := Format(AxesFormat, [q/DivNum]);
            BottomMargin := max(BottomMargin, FCanvas.TextHeight(Str));
            inc(i); q := q0 + i*gx;
          end;
        end;
      end;
      BottomMargin := max(BottomMargin, OrderHeight(Order)) + 2*AxesTextMargin;
    end;
    if ShowYnums then begin
      LeftMargin := -1;
      if LogY then begin
        Order := round(MaxY - MinY);
        DivNum := IntPower(10, Order);
        for i := round(MinY) to round(MaxY) do begin
          Str := Format(AxesFormat, [IntPower(10, i)/DivNum]);
          LeftMargin := max(LeftMargin, FCanvas.TextWidth(Str));
        end;
      end else begin
        grid_y := max(1, H div GridCellSize);
        gy := (MaxY - MinY)/grid_y;
        Order := round(log10(gy));
        gy := round(gy*IntPower(10, -Order+1))*IntPower(10, Order-1);
        DivNum := IntPower(10, Order);
        if gy > 0 then begin
          q0 := gy*int(MinY/gy);
          i := 0; q := q0;
          while q <= MaxY do begin
            if abs(q) < EPS then q := 0;
            Str := Format(AxesFormat, [q/DivNum]);
            LeftMargin := max(LeftMargin, FCanvas.TextWidth(Str));
            inc(i); q := q0 + i*gy;
          end;
        end;
      end;
      LeftMargin := max(LeftMargin, OrderWidth(Order)) + 4*AxesTextMargin;
    end;
  end;
end;

procedure TFigure.ClipCanvas(const Rect: TRect);
var
  RectRgn: HRGN;
begin
  RectRgn := CreateRectRgn(Rect.Left, Rect.Top, Rect.Right + 1, Rect.Bottom + 1);
  SelectClipRgn(FCanvas.Handle, RectRgn);
  DeleteObject(RectRgn);
end;

procedure TFigure.UnClipCanvas;
begin
  SelectClipRgn(FCanvas.Handle, 0);
end;

function TFigure.OrderWidth(const Order: integer): integer;
const
  BaseStr = 'x10';
var
  OrderStr: string;
begin
  if Order = 0 then
    result := 0
  else
    with FCanvas do begin
      result := TextWidth(BaseStr);
      if Order = 1 then exit;
      OrderStr := IntToStr(Order);
      Font.Size := Font.Size - DeltaFont;
      result := result + TextWidth(OrderStr);
      Font.Size := Font.Size + DeltaFont;
    end;
end;

function TFigure.OrderHeight(const Order: integer): integer;
const
  BaseStr = 'x10';
var
  OrderStr: string;
begin
  if Order = 0 then
    result := 0
  else
    with FCanvas do begin
      result := TextHeight(BaseStr);
      if Order = 1 then exit;
      OrderStr := IntToStr(Order);
      Font.Size := Font.Size - DeltaFont;
      result := result + TextHeight(OrderStr) div 2;
      Font.Size := Font.Size + DeltaFont;
    end;
end;

procedure TFigure.DrawOrder(const Order: integer; const P: TPoint);
const
  BaseStr = 'x10';
var
  BaseWidth, BaseHeight, OrderWidth, OrderHeight: integer;
  OrderStr: string;
  Area: TRect;
begin
  if Order <> 0 then
    with FCanvas do begin
      OrderStr := IntToStr(Order);
      BaseWidth := TextWidth(BaseStr);
      BaseHeight := TextHeight(BaseStr);
      Font.Size := Font.Size - DeltaFont;
      OrderWidth := TextWidth(OrderStr);
      OrderHeight := TextHeight(OrderStr) div 2;
      Font.Size := Font.Size + DeltaFont;

      Area := Classes.Rect(
        P.X - OrderWidth - BaseWidth + 1, P.Y,
        P.X + 1, P.Y + OrderHeight + BaseHeight);
      Brush.Color := clUnder;
      FillRect(Area);

      if Order = 1 then
        TextOut(P.X - BaseWidth + 1, P.Y, BaseStr)
      else begin
        // TextHieght returns number that larger then real digits height
        // that's why I'm using div 2
        TextOut(
          P.X - OrderWidth - BaseWidth + 1,
          P.Y + OrderHeight, BaseStr);
        Font.Size := Font.Size - DeltaFont;
        TextOut(P.X - OrderWidth + 1, P.Y, OrderStr);
        Font.Size := Font.Size + DeltaFont;
      end;
    end;
end;

procedure TFigure.Draw;
var
  i, j, k, cnt, Order, W, H, tx, ty, grid_x, grid_y, pd2, maxidx: integer;
  clLineIndex, clFillIndex, clPointIndex: integer;
  LineWidthIndex, PointSizeIndex: integer;
  CoefX, CoefY, gx, gy, q, q0, DivNum, xc, yc: double;
  Origin, XOrder, YOrder: TPoint;
  Poly: array of TPoint;
  Rect, Rt: TRect;
  Str: string;

  procedure ToPix(const X, Y: double; var PixX, PixY: integer);
  begin
    PixX := Origin.X + round(X*CoefX);
    PixY := Origin.Y - round(Y*CoefY);
  end;

begin
  Rect := FRect;
  with FCanvas do begin
    Font.Name := FontName;
    Font.Size := FontSize;
    Font.Color := clAxis;
    Pen.Width := 1;
    Pen.Style := psSolid;
    Brush.Style := bsSolid;

    Brush.Color := clBack;
    FillRect(Classes.Rect(Rect.Left, Rect.Top, Rect.Right + 1, Rect.Bottom + 1));

    if ShowTitle then begin
      TitleHeight := TextHeight(Title);
      Brush.Color := clUnder;
      Rt := Classes.Rect(
        Rect.Left + Margin, Rect.Top + Margin, Rect.Right - Margin + 1,
        Rect.Top + TitleHeight + Margin);
      FillRect(Rt);
      TitleVisible := TextWidth(Title) <= Rt.Right - Rt.Left + 1 -
        2*TitleTextMargin;
      if TitleVisible then Str := Title
      else Str := '...';
      TextOut(
        (Rect.Left + Rect.right - TextWidth(Str)) div 2 + 1,
        Rect.Top + Margin, Str);
      Rect.Top := Rect.Top + TitleHeight + Margin div 2;
    end;

    W := Rect.right - Rect.left + 1 - 3*Margin - LeftMargin;
    H := Rect.bottom - Rect.top + 1 - 3*Margin - BottomMargin;
    if (W <= 1) or (H <= 1) then exit;

    Rt := Classes.Rect(
      Rect.Left + Margin, Rect.Top + Margin,
      Rect.Right - Margin + 1, Rect.Bottom - Margin + 1);

    Brush.Color := clUnder;
    FillRect(Rt);

    if (N > 0) and (MaxX > NegInfinity) and (MinX < Infinity) and
       (MaxY > NegInfinity) and (MinY < Infinity) then begin
      CoefX := W/abs(MaxX - MinX);
      CoefY := H/abs(MaxY - MinY);

      Origin.X := Rect.Left + Margin + LeftMargin - round(MinX*CoefX);
      Origin.Y := Rect.Top + 2*Margin + round(MaxY*CoefY);

      ClipCanvas(Classes.Rect(
        Rt.Left, Rt.Top + Margin, Rt.Right - Margin - 1, Rt.Bottom - 1));

      Pen.Color := clGrid;

      XOrder.X := Rect.Right - 2*Margin;
      XOrder.Y := Rect.Bottom - Margin - BottomMargin + 1 + AxesTextMargin;

      YOrder.X := Rect.Left + Margin + LeftMargin - 2 - AxesTextMargin;
      YOrder.Y := Rect.Top + 2*Margin;

      if LogX then begin
        Order := round(MaxX - MinX);
        DivNum := IntPower(10, Order);
        for i := round(MinX) to round(MaxX) do begin
          ToPix(i, MinY, tx, ty);
          if (tx >= Rect.Left + Margin + LeftMargin) and
             (tx <= Rect.Right - 2*Margin) then begin
            if ShowYgrid then begin
              MoveTo(tx, Rect.Bottom - Margin - BottomMargin);
              LineTo(tx, Rect.top + 2*Margin - 1);
              for j := 1 to 10 do begin
                ToPix(i + log10(j), MinY, tx, ty);
                if (tx >= Rect.Left + Margin + LeftMargin) and
                   (tx <= Rect.Right - 2*Margin) then begin
                  MoveTo(tx, Rect.bottom - Margin - BottomMargin);
                  LineTo(tx, Rect.Top + 2*Margin - 1);
                end;
              end;
            end;
            if ShowXnums then begin
              ToPix(i, MinY, tx, ty);
              Str := Format(AxesFormat, [IntPower(10, i)/DivNum]);
              if tx + TextWidth(Str) div 2 <= Rt.Right - 1 then
                TextOut(
                  tx - TextWidth(Str) div 2,
                  XOrder.Y - 1,
                  Str);
            end;
          end;
        end;
        if ShowXnums then
          DrawOrder(Order, XOrder);
      end else begin
        grid_x := max(1, W div GridCellSize);
        gx := (MaxX - MinX)/grid_x;
        Order := round(log10(gx));
        gx := round(gx*IntPower(10, -Order+1))*IntPower(10, Order-1);
        DivNum := IntPower(10, Order);
        if gx > 0 then begin
          q0 := gx*int(MinX/gx);
          i := 0; q := q0;
          while q <= MaxX do begin
            ToPix(q, MinY, tx, ty);
            if (tx >= Rect.Left + Margin + LeftMargin) and
               (tx <= Rect.Right - 2*Margin) then begin
              if ShowYgrid then begin
                MoveTo(tx, Rect.Bottom - Margin - BottomMargin);
                LineTo(tx, Rect.Top + 2*Margin - 1);
              end;
              if abs(q) < EPS then q := 0;
              if ShowXnums then begin
                Str := Format(AxesFormat, [q/DivNum]);
                if tx + TextWidth(Str) div 2 <= Rt.Right - 1 then
                  TextOut(
                    tx - TextWidth(Str) div 2,
                    XOrder.Y - 1,
                    Str);
              end;
            end;
            inc(i); q := q0 + i*gx;
          end;
          if ShowXnums then
            DrawOrder(Order, XOrder);
        end;
      end;

      if LogY then begin
        Order := round(MaxY - MinY);
        DivNum := IntPower(10, Order);
        for i := round(MinY) to round(MaxY) do begin
          ToPix(MinX, i, tx, ty);
          if (ty >= Rect.Top + 2*Margin) and
             (ty <= Rect.Bottom - Margin - BottomMargin) then begin
            if ShowXgrid then begin
              MoveTo(Rect.Left + Margin + LeftMargin, ty);
              LineTo(Rect.Right - 2*Margin + 1, ty);
              for j := 1 to 10 do begin
                ToPix(MinX, i + log10(j), tx, ty);
                if (ty >= Rect.Top + 2*Margin) and
                   (ty <= Rect.Bottom - Margin - BottomMargin) then begin
                  MoveTo(Rect.Left + Margin + LeftMargin, ty);
                  LineTo(Rect.Right - 2*Margin + 1, ty);
                end;
              end;
            end;
            if ShowYnums then begin
              ToPix(MinX, i, tx, ty);
              Str := Format(AxesFormat, [IntPower(10, i)/DivNum]);
              if ty - TextHeight(Str) div 2 >= Rt.Top then
                TextOut(
                  YOrder.X - TextWidth(Str),
                  ty - TextHeight(Str) div 2,
                  Str);
            end;
          end;
        end;
        if ShowYnums then
          DrawOrder(Order, YOrder);
      end else begin
        grid_y := max(1, H div GridCellSize);
        gy := (MaxY - MinY)/grid_y;
        Order := round(log10(gy));
        gy := round(gy*IntPower(10, -Order+1))*IntPower(10, Order-1);
        DivNum := IntPower(10, Order);
        if gy > 0 then begin
          q0 := gy*int(MinY/gy);
          i := 0; q := q0;
          while q <= MaxY do begin
            ToPix(MinX, q, tx, ty);
            if (ty >= Rect.Top + 2*Margin) and
               (ty <= Rect.Bottom - Margin - BottomMargin) then begin
              if ShowXgrid then begin
                MoveTo(Rect.Left + Margin + LeftMargin, ty);
                LineTo(Rect.Right - 2*Margin + 1, ty);
              end;
              if abs(q) < EPS then q := 0;
              if ShowYnums then begin
                Str := Format(AxesFormat, [q/DivNum]);
                if ty - TextHeight(Str) div 2 >= Rt.Top then
                  TextOut(
                    YOrder.X - TextWidth(Str),
                    ty - TextHeight(Str) div 2,
                    Str);
              end;
            end;
            inc(i); q := q0 + i*gy;
          end;
          if ShowYnums then
            DrawOrder(Order, YOrder);
        end;
      end;

      UnClipCanvas;

      Rt := Classes.Rect(
        Rect.Left + Margin + LeftMargin,
        Rect.Top + 2*Margin, Rect.Right - 2*Margin,
        Rect.Bottom - Margin - BottomMargin);

      ClipCanvas(Rt);

      Pen.Color := clAxis;
      MoveTo(Rect.Left + Margin + LeftMargin, Origin.Y);
      LineTo(Rect.Right - 2*Margin + 1, Origin.Y);
      MoveTo(Origin.X, Rect.Top + 2*Margin);
      LineTo(Origin.X, Rect.Bottom - Margin - BottomMargin + 1);

      clLineIndex := 0;
      clFillIndex := 0;
      clPointIndex := 0;
      LineWidthIndex := 0;
      PointSizeIndex := 0;
      for i := 0 to N-1 do
        if Length(X[i]) > 0 then begin
          Pen.Width := LineWidth[LineWidthIndex];
          if Pen.Width = 0 then Pen.Style := psClear
          else Pen.Style := psSolid;
          Pen.Color := clLine[clLineIndex];
          Brush.Color := clPoint[clPointIndex];

          maxidx := High(X[i]);
          SetLength(Poly, maxidx + 1);
          for j := 0 to maxidx do begin
            if LogX and (X[i,j] > 0) then xc := log10(X[i,j]) else xc := X[i,j];
            if LogY and (Y[i,j] > 0) then yc := log10(Y[i,j]) else yc := Y[i,j];
            ToPix(xc, yc, Poly[j].x, Poly[j].y);
          end;
          if X[i,0] = X[i,maxidx] then begin
            Brush.Color := clFill[clFillIndex];
            Polygon(Poly);
            Brush.Color := clPoint[clPointIndex];
          end else
            Polyline(Poly);

          pd2 := PointSize[PointSizeIndex] div 2;
          for j := 0 to maxidx do
            Ellipse(
              Poly[j].x - pd2, Poly[j].y - pd2,
              Poly[j].x + pd2, Poly[j].y + pd2);

          clLineIndex := (clLineIndex + 1) mod length(clLine);
          clFillIndex := (clFillIndex + 1) mod length(clFill);
          clPointIndex := (clPointIndex + 1) mod length(clPoint);
          LineWidthIndex := (LineWidthIndex + 1) mod length(LineWidth);
          PointSizeIndex := (PointSizeIndex + 1) mod length(PointSize);
        end;

      UnClipCanvas;
    end;
  end;
end;

procedure TFigure.SaveToBMP(const FileName: string;
  const Width, Height: integer);
var
  Bmp: TBitmap;
  OldCanvas: TCanvas;
  OldRect: TRect;
begin
  Bmp := TBitmap.Create;
  try
    Bmp.Width := Width;
    Bmp.Height := Height;

    OldCanvas := FCanvas; FCanvas := Bmp.Canvas;
    OldRect := FRect; FRect := Rect(0, 0, Width-1, Height-1);
    Draw;
    FCanvas := OldCanvas;
    FRect := OldRect;

    Bmp.SaveToFile(FileName);
  finally
    Bmp.Free;
  end;
end;

procedure TFigure.SaveToJPG(const FileName: string;
  const Width, Height: integer);
var
  Bmp: TBitmap;
  Jpg: TJPEGImage;
  OldCanvas: TCanvas;
  OldRect: TRect;
begin
  Bmp := TBitmap.Create;
  Jpg := TJPEGImage.Create;
  try
    Bmp.Width := Width;
    Bmp.Height := Height;

    OldCanvas := FCanvas; FCanvas := Bmp.Canvas;
    OldRect := FRect; FRect := Rect(0, 0, Width-1, Height-1);
    Draw;
    FCanvas := OldCanvas;
    FRect := OldRect;

    Jpg.Assign(Bmp);
    Jpg.SaveToFile(FileName);
  finally
    Jpg.Free;
    Bmp.Free;
  end;
end;

procedure TFigure.SaveToEMF(const FileName: string;
  const Width, Height: integer);
var
  Emf: TMetaFile;
  Canvas: TMetaFileCanvas;
  OldCanvas: TCanvas;
  OldRect: TRect;
begin
  Emf := TMetaFile.Create;
  try
    Emf.Width := Width - 1;
    Emf.Height := Height - 1;

    Canvas := TMetaFileCanvas.Create(Emf, Emf.Handle);
    try
      OldCanvas := FCanvas; FCanvas := Canvas;
      OldRect := FRect; FRect := Rect(0, 0, Width-1, Height-1);
      Draw;
      FCanvas := OldCanvas;
      FRect := OldRect;
    finally
      Canvas.Free;
    end;

    Emf.SaveToFile(FileName);
  finally
    Emf.Free;
  end;
end;

procedure TFigure.SaveToFile(const FileName: string;
  const Width, Height: integer);
var
  ext: string;
begin
  ext := LowerCase(ExtractFileExt(FileName));
  if ext = '.bmp' then SaveToBMP(FileName, Width, Height)
  else if (ext = '.jpg') or (ext = '.jpeg') then SaveToJPG(FileName, Width, Height)
  else if ext = '.emf' then SaveToEMF(FileName, Width, Height);
end;

// TFigureWindow ///////////////////////////////////////////////////////////////

function TFigureWindow.GetFigure(FigureNumber: integer): TFigure;
begin
  if (FigureNumber < 0) or (FigureNumber > high(Figure)) then
    result := nil
  else
    result := Figure[FigureNumber];
end;

procedure TFigureWindow.SetCanvas(const NewCanvas: TCanvas);
var
  i: integer;
begin
  FCanvas := NewCanvas;
  for i := 0 to High(Figure) do
    Figure[i].Canvas := NewCanvas;
end;

procedure TFigureWindow.Resize(const NewWidth, NewHeight: integer);
var
  i: integer;
begin
  FWidth := NewWidth;
  FHeight := NewHeight;
  for i := 0 to High(Figure) do
    Figure[i].Resize(GetRect(i));
end;

procedure TFigureWindow.HomeFigure;
begin
  StartFig := BegFig;
  Load(false);
end;

procedure TFigureWindow.EndFigure;
begin
  StartFig := EndFig - WFig*HFig + 1;
  Load(false);
end;

function TFigureWindow.NextFigure(const One: boolean): boolean;
var
  k: integer;
begin
  if One then k := 1 else k := WFig*HFig;
  result := StartFig + WFig*HFig <= EndFig;
  if result then begin
    Inc(StartFig, k);
    Load(false);
  end;
end;

function TFigureWindow.PrevFigure(const One: boolean): boolean;
begin
  result := false;
  if One then begin
    if StartFig > BegFig then begin
      Dec(StartFig);
      result := true;
    end;
  end else begin
    if StartFig - 1 >= BegFig then begin
      Dec(StartFig, WFig*HFig);
      result := true;
    end;
  end;
  if result then
    Load(false);
end;

procedure TFigureWindow.MoreFigure(const Xs: boolean);
begin
  if Xs then Inc(WFig) else Inc(HFig);
  Load(false);
end;

function TFigureWindow.LessFigure(const Xs: boolean): boolean;
begin
  result := false;
  if Xs then begin
    if WFig > 1 then begin
      Dec(WFig);
      result := true;
    end;
  end else begin
    if HFig > 1 then begin
      Dec(HFig);
      result := true;
    end;
  end;
  if result then
    Load(false);
end;

constructor TFigureWindow.Create(const FileName: string);
begin
  WFig := 0; HFig := 0;
  SetLength(Figure, 0);
  Ini := TMemIniFile.Create(FileName);
end;

destructor TFigureWindow.Destroy;
var
  i: integer;
begin
  if Ini <> nil then Ini.Free;
  for i := 0 to high(Figure) do
    if Figure[i] <> nil then
      FreeAndNil(Figure[i]);
end;

procedure TFigureWindow.Load(const ReadSizes: boolean = true);
var
  dclAxis, dclBack, dclUnder, dclGrid: TColor;
  dmargin_size: integer;
  dshow_title, dshow_Xgrid, dshow_Ygrid, dlogX, dlogY: boolean;
  dshow_Xnums, dshow_Ynums: boolean;
  dminx, dmaxx, dminy, dmaxy, kX, kY, InnerMargin: double;
  I, J, nf, dfontsize, c, xb, min_width, min_height: integer;
  dclPoint, dlwidth, dpoint_size, dfontname, dclLine, dclFill, ss, ns: string;
  calcmin, calcmax: boolean;
  Rt: TRect;

  procedure StrToArray(const Str: string; var Arr: TArrayOfInteger;
    const DefaultValue: integer = 0); overload;
  var
    i, cnt: integer;
  begin
    i := 1; cnt := 0;
    while i <= length(Str) do begin
      StrTok(Str, i, SpaceChars);
      inc(cnt);
    end;
    SetLength(Arr, cnt);
    i := 1; cnt := 0;
    while i <= length(Str) do begin
      try
        Arr[cnt] := StrToInt(StrTok(Str, i, SpaceChars));
      except
        Arr[cnt] := DefaultValue;
      end;
      inc(cnt);
    end;
  end;

  procedure StrToArray(const Str: string; var Arr: TArrayOfDouble;
    const DefaultValue: double = 0); overload;
  var
    i, cnt: integer;
  begin
    i := 1; cnt := 0;
    while i <= length(Str) do begin
      StrTok(Str, i, SpaceChars);
      inc(cnt);
    end;
    SetLength(Arr, cnt);
    i := 1; cnt := 0;
    while i <= length(Str) do begin
      try
        Arr[cnt] := StrToFloat(StrTok(Str, i, SpaceChars));
      except
        Arr[cnt] := DefaultValue;
      end;
      inc(cnt);
    end;
  end;

  procedure TestRange(var Value: integer;
    const MinValue, MaxValue: integer); inline; overload;
  begin
    Value := max(min(Value, MaxValue), MinValue);
  end;

  procedure TestRange(var Arr: array of integer;
    const MinValue, MaxValue: integer); overload;
  var i: integer;
  begin
    for i := Low(Arr) to High(Arr) do
      Arr[i] := max(min(Arr[i], MaxValue), MinValue);
  end;

begin
  try
    // [Window]
    Cap := Trim(ini.ReadString('Window', 'Caption', ''));
    PlottoFile := ini.ReadString('Window', 'PlotToFile', '');
    BegFig := ini.ReadInteger('Window', 'BegFig', 1);
    TestRange(BegFig, 1, MaxInt);
    EndFig := ini.ReadInteger('Window', 'EndFig', BegFig);
    TestRange(EndFig, 1, MaxInt);
    if ReadSizes then begin
      WFig := ini.ReadInteger('Window', 'WFig', 1);
      TestRange(WFig, 1, MaxInt);
      HFig := ini.ReadInteger('Window', 'HFig', 1);
      TestRange(HFig, 1, MaxInt);
      StartFig := BegFig;
    end;

    min_width := FigureConstraint*WFig;
    min_height := FigureConstraint*HFig;

    if ReadSizes then begin
      I := ini.ReadInteger('Window', 'Width', 400);
      FWidth := max(I, min_width);
      I := ini.ReadInteger('Window', 'Height', 300);
      FHeight := max(I, min_height);
    end;

    // [Defaults]
    dclLine  := ini.ReadString('Defaults', 'clLine',
      '$000000 $000088 $880088 $008800');
    dclFill  := ini.ReadString('Defaults', 'clFill',
      '$FFC8AA $AAC8FF $FFB4FF $A0FFCD');
    dclBack  := ini.ReadInteger('Defaults', 'clBack',  $888888);
    dclUnder := ini.ReadInteger('Defaults', 'clUnder', $EEEEEE);
    dclPoint := ini.ReadString('Defaults', 'clPoint', '$0000FF');
    dclAxis  := ini.ReadInteger('Defaults', 'clAxis',  $880000);
    dclGrid  := ini.ReadInteger('Defaults', 'clGrid',  $CCCCCC);

    dfontname := ini.ReadString('Defaults', 'FontName', 'Tahoma');
    dfontsize := ini.ReadInteger('Defaults', 'FontSize', 8);

    dlwidth := ini.ReadString('Defaults', 'LineWidth', '1');
    dpoint_size := ini.ReadString('Defaults', 'PointSize', '5');
    dmargin_size := ini.ReadInteger('Defaults', 'Margin', 5);
    dshow_Xnums := ini.ReadBool('Defaults', 'ShowTicksX', true);
    dshow_Ynums := ini.ReadBool('Defaults', 'ShowTicksY', true);

    dshow_title := ini.ReadBool('Defaults', 'ShowTitle', false);
    dshow_Xgrid := ini.ReadBool('Defaults', 'ShowGridX', true);
    dshow_Ygrid := ini.ReadBool('Defaults', 'ShowGridY', true);
    dlogX := ini.ReadBool('Defaults', 'LogX', false);
    dlogY := ini.ReadBool('Defaults', 'LogY', false);

    dminx := ini.ReadFloat('Defaults', 'MinX', Infinity);
    dmaxx := ini.ReadFloat('Defaults', 'MaxX', NegInfinity);
    dminy := ini.ReadFloat('Defaults', 'MinY', Infinity);
    dmaxy := ini.ReadFloat('Defaults', 'MaxY', NegInfinity);

    nf := WFig*HFig;
    setlength(Figure, nf);

    // [FigureXX]
    for I := 0 to nf-1 do begin
      Figure[I] := TFigure.Create(I);
      if (StartFig + I < BegFig) or (StartFig + I > EndFig) then begin
        with Figure[I] do begin
          StrToArray(dclLine, clLine, clBlack);
          clBack  := dclBack;
          clUnder := dclUnder;
          StrToArray(dclPoint, clPoint, clRed);
          clAxis  := dclAxis;
          clGrid  := dclGrid;

          StrToArray(dlwidth, LineWidth);
          StrToArray(dpoint_size, PointSize);
          Margin := dmargin_size;

          FontName := dfontname;
          FontSize := dfontsize;

          ShowTitle := dshow_title;
          ShowXgrid := dshow_Xgrid;
          ShowYgrid := dshow_Ygrid;
          LogX := dlogX;
          LogY := dlogY;

          SMinX := dminx; SMaxX := dmaxx;
          SMinY := dminy; SMaxY := dmaxy;

          Title := ' ';
          TitleHeight := 0;
        end;
        Figure[I].Canvas := FCanvas;
        Figure[I].Resize(GetRect(I));
        continue;
      end;

      ss := 'Figure' + IntToStr(StartFig + I);
      with Figure[I] do begin
        StrToArray(ini.ReadString(ss, 'clLine', dclLine), clLine, clBlack);
        StrToArray(ini.ReadString(ss, 'clFill', dclFill), clFill, clWhite);
        clBack  := ini.ReadInteger(ss, 'clBack',  dclBack);
        clUnder := ini.ReadInteger(ss, 'clUnder', dclUnder);
        StrToArray(ini.ReadString(ss, 'clPoint', dclPoint), clPoint, clRed);
        clAxis  := ini.ReadInteger(ss, 'clAxis',  dclAxis);
        clGrid  := ini.ReadInteger(ss, 'clGrid',  dclGrid);

        StrToArray(ini.ReadString(ss, 'LineWidth', dlwidth), LineWidth);
        TestRange(LineWidth, 0, MaxInt);
        StrToArray(ini.ReadString(ss, 'PointSize', dpoint_size), PointSize);
        TestRange(PointSize, 0, MaxInt);
        Margin := ini.ReadInteger(ss, 'Margin', dmargin_size);
        TestRange(Margin, 0, MaxInt);
        ShowXnums := ini.ReadBool(ss, 'ShowTicksX', dshow_Xnums);
        ShowYnums := ini.ReadBool(ss, 'ShowTicksY', dshow_Ynums);
        
        FontName := ini.ReadString(ss, 'FontName', dfontname);
        FontSize := ini.ReadInteger(ss, 'FontSize', dfontsize);
        TestRange(FontSize, 4, MaxInt);

        ShowTitle := ini.ReadBool(ss, 'ShowTitle', dshow_title);
        ShowXgrid := ini.ReadBool(ss, 'ShowGridX', dshow_Xgrid);
        ShowYgrid := ini.ReadBool(ss, 'ShowGridY', dshow_Ygrid);
        LogX := ini.ReadBool(ss, 'LogX', dlogX);
        LogY := ini.ReadBool(ss, 'LogY', dlogY);

        Title := ini.ReadString(ss, 'Title', ' ');

        N := ini.ReadInteger(ss, 'DataCount', 0);
        setlength(X, N, 0);
        setlength(Y, N, 0);
        if N > 0 then begin
          for J := 0 to N-1 do begin
            ns := IntToStr(J+1);
            StrToArray(ini.ReadString(ss, 'X' + ns, ''), Figure[I].X[J]);
            StrToArray(ini.ReadString(ss, 'Y' + ns, ''), Figure[I].Y[J]);
          end;
        end;

        SMinX := ini.ReadFloat(ss, 'MinX', dminx);
        SMaxX := ini.ReadFloat(ss, 'MaxX', dmaxx);
        SMinY := ini.ReadFloat(ss, 'MinY', dminy);
        SMaxY := ini.ReadFloat(ss, 'MaxY', dmaxy);
        if SMinX >= SMaxX then begin
          calcmin := SMinX = Infinity;
          calcmax := SMaxX = NegInfinity;
          for J := 0 to N-1 do
            if length(X[J]) > 0 then begin
              if calcmin then SMinX := min(SMinX, MinValue(X[J]));
              if calcmax then SMaxX := max(SMaxX, MaxValue(X[J]));
            end;
        end;
        if LogX then begin
          if SMinX > 0 then SMinX := log10(SMinX);
          if SMaxX > 0 then SMaxX := log10(SMaxX);
          if SMinX > SMaxX then swap(SMinX, SMaxX);
        end;
        SMinX := SMinX - EPS;
        SMaxX := SMaxX + EPS;
        if SMinY >= SMaxY then begin
          calcmin := SMinY = Infinity;
          calcmax := SMaxY = NegInfinity;
          for J := 0 to N-1 do
            if length(Y[J]) > 0 then begin
              if calcmin then SMinY := min(SMinY, MinValue(Y[J]));
              if calcmax then SMaxY := max(SMaxY, MaxValue(Y[J]));
            end;
        end;
        if LogY then begin
          if SMinY > 0 then SMinY := log10(SMinY);
          if SMaxY > 0 then SMaxY := log10(SMaxY);
          if SMinY > SMaxY then swap(SMinY, SMaxY);
        end;
        SMinY := SMinY - EPS;
        SMaxY := SMaxY + EPS;
      end;
      Figure[I].Canvas := FCanvas;
      Figure[I].Resize(GetRect(I));
    end; // for all figures
  except
    Application.MessageBox(PAnsiChar('Error while reading options from "' +
      OptionsFile + '"!'), 'Error');
  end;
end;

procedure TFigureWindow.Draw;
var
  i: integer;
begin
  for i := 0 to high(Figure) do
    Figure[i].Draw;
end;

function TFigureWindow.Hint(const X, Y: integer; const Precision: byte;
  var HintType: integer): string;
var
  i, j, w, h: integer;
  xx, yy, kX, kY: double;
  Str: string;
  Rt: TRect;
begin
  HintType := 0;
  result := '';
  j := -1;
  for i := 0 to high(Figure) do begin
    if not Figure[i].TitleVisible and InRect(GetTitleRect(i), X, Y) then begin
      result := ' ' + Figure[i].Title + ' ';
      HintType := 1;
      exit;
    end;
    if InRect(GetGraphRect(i), X, Y) then begin
      j := i; break;
    end;
  end;
  if j < 0 then exit;

  with Figure[j] do begin
    if (N <= 0) or (MaxX = NegInfinity) or (MinX = Infinity) and
      (MaxY = NegInfinity) or (MinY = Infinity) then exit;

    Rt := GetRect(j);
    if ShowTitle then
      Rt.Top := Rt.Top + FCanvas.TextHeight(Title) + Margin div 2;
    w := Rt.Right - Rt.Left + 1;
    h := Rt.Bottom - Rt.Top + 1;

    kX := (w - 3*Margin - LeftMargin)/abs(MaxX - MinX);
    kY := (h - 3*Margin - BottomMargin)/abs(MaxY - MinY);

    xx := Rt.Left + Margin + LeftMargin - round(MinX*kX);
    yy := Rt.Top + 2*Margin + round(MaxY*kY);
  end;
  xx := (X - xx)/kX;
  yy := (yy - Y)/kY;
  if Figure[j].LogX then xx := Power(10, xx);
  if Figure[j].LogY then yy := Power(10, yy);

  Str := '%0.' + IntToStr(Precision) + 'f';
  result := Format(' X: ' + Str + ' Y: ' + Str + ' ', [xx, yy]);
  HintType := 2;
end;

procedure TFigureWindow.SaveToFile(const FileName: string);
var
  ext: string;
  Bmp: TBitmap;
  Jpg: TJPEGImage;
  Emf: TMetaFile;
  Canvas: TMetaFileCanvas;
  OldCanvas: TCanvas;
begin
  ext := LowerCase(ExtractFileExt(FileName));
  if ext = '.bmp' then begin
    Bmp := TBitmap.Create;
    try
      Bmp.Width := FWidth;
      Bmp.Height := FHeight;

      OldCanvas := FCanvas;
      SetCanvas(Bmp.Canvas);
      Draw;
      FCanvas := OldCanvas;

      Bmp.SaveToFile(FileName);
    finally
      Bmp.Free;
    end;
  end else if (ext = '.jpg') or (ext = '.jpeg') then begin
    Bmp := TBitmap.Create;
    Jpg := TJPEGImage.Create;
    try
      Bmp.Width := FWidth;
      Bmp.Height := FHeight;

      OldCanvas := FCanvas;
      SetCanvas(Bmp.Canvas);
      Draw;
      FCanvas := OldCanvas;

      Jpg.Assign(Bmp);
      Jpg.SaveToFile(FileName);
    finally
      Jpg.Free;
      Bmp.Free;
    end;
  end else if ext = '.emf' then begin
    Emf := TMetaFile.Create;
    try
      Emf.Width := FWidth - 1;
      Emf.Height := FHeight - 1;

      Canvas := TMetaFileCanvas.Create(Emf, Emf.Handle);
      try
        OldCanvas := FCanvas;
        SetCanvas(Canvas);
        Draw;
        FCanvas := OldCanvas;
      finally
        Canvas.Free;
      end;

      Emf.SaveToFile(FileName);
    finally
      Emf.Free;
    end;
  end;
end;

function TFigureWindow.WhatFigure(const X, Y: integer): integer;
var
  i: integer;
begin
  result := -1;
  for i := 0 to High(Figure) do
    if InRect(GetRect(i), X, Y) then begin
      result := i; exit;
    end;
end;

function TFigureWindow.GetRect(const FigureNumber: integer): TRect;
var
  i, j, w, h: integer;
begin
  w := FWidth div WFig;
  h := FHeight div HFig;
  i := FigureNumber div WFig;
  j := FigureNumber mod WFig;
  result.Left   := j*w;
  result.Right  := (j+1)*w - 1;
  result.Top    := i*h;
  result.Bottom := (i+1)*h - 1;
end;

function TFigureWindow.GetTitleRect(const FigureNumber: integer): TRect;
begin
  with Figure[FigureNumber] do begin
    result := GetRect(FigureNumber);
    result.Left := result.Left + Margin;
    result.Top := result.Top + Margin;
    result.Right := result.Right - Margin;
    result.Bottom := result.Top + TitleHeight;
  end;
end;

function TFigureWindow.GetGraphRect(const FigureNumber: integer): TRect;
begin
  with Figure[FigureNumber] do begin
    result := GetRect(FigureNumber);
    result.Left := result.Left + LeftMargin + Margin;
    result.Top := result.Top + TitleHeight + 2*Margin +
      ord(TitleHeight <> 0)*(Margin div 2);
    result.Right := result.Right - 2*Margin;
    result.Bottom := result.Bottom - BottomMargin - Margin;
  end;
end;

end.